
source('functions.R')

load("MFCC-aast.RData")
load(sprintf("locations.RData"))
library(MASS)
library(automap)
library(dplyr)
library(fields)
require(tuneR)
require(audio)
require(akima)
require(ggmap)
library(ggplot2)
library(sp)
library(maps)
library(gstat)
library(fdaPDE)
library(rworldmap)
library(rworldxtra)


##### settings #####

showplot <- FALSE  # plot (TRUE) or not (FALSE) location maps
compute.cv <- FALSE # compute (TRUE) or not (FALSE) cross-validation errors
cv.plots <- FALSE
compute.smooth.field <- TRUE ## compute (TRUE) or not (FALSE) the smooth field with the tuning parameters specified below
show.pc.plots <- FALSE  # plot (TRUE) or not (FALSE) the colormap of the first 3 pc scores and loadings.
sqrt.maps.compute <- FALSE # compute (TRUE - Warning: this is slow) or not (FALSE) the distance between d-covariance  with square-root metric, in a set of location and in the rest of the region 
l2.maps.compute <- FALSE # compute (TRUE - Warning: this is slow) or not (FALSE) the distance between d-covariance, with L2 metric, in a set of location and in the rest of the region 
sqrt.vs.l2.maps <- FALSE # plot maps of distances of d-cov from a location, with square-root metric vs L2 metric
geodesic.vs.euclidean.distance <- FALSE # plot pc projections of the mean field obtained with geodesic geographical distances vs euclidean (straight line) geographical distances 
constant.field.simulation  <- FALSE # simulate data from a constant field (constant mean and constant distribution of the errors)
constant.field.pc.plots  <- FALSE # pc plots for mean field of simulation
plot.sim.contours <- FALSE # contour plots of simulations and BNC data
compute.zscores <- FALSE # compute zscores
plot.zscores <- FALSE # plot zscore maps
distance.vs.residual.compute <- FALSE # compute pairwise distances between residuals
distance.vs.residual.plot <- FALSE # plot pairwise distances between residuals vs the geographical distance
covdist.vs.residual.compute <- FALSE # plot of pairwise distances between raw covariances vs the geographical distance
covdist.vs.residual.plot <- FALSE # plot of pairwise distances between raw covariances vs the geographical distance
duration.field.plot <- FALSE # model the relative vowel duration using a linear mixed model, and plot the relative vowel duration field





### name of the dataset ###
data.name <- 'aast'

### cross-validation parameters ###

h.seq <- c(0.3,0.4,seq(0.5, 3.3, by=.5))
h.seq.sqrt <- c(0.3,0.4,seq(0.5, 3.4, by=.5))
NN.location.seq <- seq(8, 32, by=6)
NN.effective.seq <- seq(200, 600, by=100)

### mfcc parameters ###

tsub<-1:20
csub <- 1:10
mfcc.data<-get(paste0(data.name, '.vowels'))$mfcc[, tsub, csub]
mfcc.data[,,1] <- sweep(mfcc.data[,,1], 1, -mean(mfcc.data[,,1]) + rowMeans(mfcc.data[, , 1]))  # renormalizing cepstral coefficient 1 to have the same average everywhere
local.bnc <- get(paste0(data.name, '.vowels.bnc'))[, ]
cov.smooth.type='sqrt'
fit.degree<-0

#### graphical parameters ####

res=200
res.cex=2.4
nlevel <- 64
field.cols <- rev(heat.colors(64))
mar1 <- c(7.5, 2.3, 4, 0.4)
.pardefault <- par(no.readonly = T)
nlevel.col <- 64
my.cols <- rev(heat.colors(nlevel.col))

### definition of mesh border  ####
data(countriesHigh)
uk.id <- grep("Kingdom", countriesHigh$NAME)
ireland.id <- grep("Ireland", countriesHigh$NAME)
uk.nodes.full <- countriesHigh@polygons[[uk.id]]@Polygons[[1]]@coords
uk.nodes <- uk.nodes.full <- uk.nodes.full[-nrow(uk.nodes.full), ] # remove last node (duplicated)
uk.segments <- outer(0:nrow(uk.nodes), 0:1, '+') %% nrow(uk.nodes)  + 1
### values == coarser approximation
eps <- .01 # value of .01 creates a mesh that contains all the data
plot(uk.nodes.full, type='l')
library(INLA) 
idx <- inla.simplify.curve(loc=uk.nodes.full, idx=1:nrow(uk.nodes.full), eps=eps)
print(c(nrow(uk.nodes.full), length(idx)))
lines(uk.nodes.full[idx, ], col='red')
border.nodes <- uk.nodes.full[idx,]


### link mfcc with mesh locations ###

prepared.data <- prepare_mfccs(mfcc.data, local.bnc, border.nodes, min.angle=30)

### list of location for which the d-covariance distance should be plotted ###
location.names <- c("devon: newton abbot, uk" , "leicestershire: leicester, uk" , "west midlands: birmingham, uk" , "avon: bristol, uk" , "berkshire: sandhurst, uk" , "hampshire: andover, uk" , "staffordshire: newcastle under lyme, uk" , "essex: harlow, uk" , "kent: bickley, uk" , "wales, holywell, uk" , "hereford & worcester: redditch, uk" , "glasgow, uk" , "strathclyde: kilmarnock, uk" , "lancashire: morecambe, uk" , "norfolk: walpole st peter, uk" , "suffolk: ipswich, uk" , "greater london: westminster, uk" , "driffield, uk" , "dorset: poole, uk" , "lothian: edinburgh, uk" , "lothian: inverness, uk" , "south lanarkshire: lanark, uk" , "yorkshire, uk" , "shropshire, uk" , "northumberland: newcastle, uk" , "lincolnshire: lincoln, uk" , "northamptonshire: irthlingborough, uk" , "borders: galashiels, uk")
location.list <- which(row.names(prepared.data$loc.coords) %in% location.names)

### location plot ###

if(showplot){
    xlim=c(-3, -2)
    ylim=c(53, 54.5)
    xlim=NULL; ylim=NULL
    plot(prepared.data$mesh$nodes,col=0,xaxt="n",yaxt="n",xlab="",ylab="",bty="n",asp=1)
    points(prepared.data$loc.coords,col=4,lwd=1,pch=3)
    lines(uk.nodes[ ,],  col='black',lwd=2)
}

### cross-validation error ###
if(compute.cv)
{
    mean.cv <- list()
    cov.sqrt.cv <- list()
    band.modes <- c('location.NN', 'effective.NN')

    
    for(band.mode in band.modes)
        {
            print(sprintf('********************  %s  ********************', band.mode))
            NN.seq <- switch(band.mode, location.NN = NN.location.seq,
                             effective.NN = NN.effective.seq, fixed = NA)


            mean.cv[[paste0(band.mode, '.deg', fit.degree)]] <- smooth_cv(what='mean', mfcc.data, local.bnc, border.nodes, min.angle=20, h.seq=h.seq, NN.seq=NN.seq, bandwidth.mode=band.mode, fit.degree=fit.degree)
            
            cov.sqrt.cv[[paste0(band.mode, '.deg', fit.degree)]] <- smooth_cv(what='cov.sqrt', mfcc.data, local.bnc, border.nodes, min.angle=20, h.seq=h.seq.sqrt, NN.seq=NN.seq, bandwidth.mode=band.mode, fit.degree=fit.degree)


        }
    save(mean.cv, cov.sqrt.cv,  file=paste0("cross-validation-", data.name, ".RData"))
}


### plot of cross-validation error curves ###

if(cv.plots)
{   
  if(file.exists(paste0("cross-validation-", data.name, ".RData"))){
    load(paste0("cross-validation-", data.name, ".RData"))
  }
    cov.min<-min(c(colMeans(cov.sqrt.cv$location.NN.deg0$cv),colMeans(cov.sqrt.cv$effective.NN.deg0$cv)))
    cov.max<-max(c(colMeans(cov.sqrt.cv$location.NN.deg0$cv),colMeans(cov.sqrt.cv$effective.NN.deg0$cv)))
    mean.min<-min(c(colMeans(mean.cv$location.NN.deg0$cv),colMeans(mean.cv$effective.NN.deg0$cv)))
    mean.max<-max(c(colMeans(mean.cv$location.NN.deg0$cv),colMeans(mean.cv$effective.NN.deg0$cv)))
    for(what in c('mean', 'cov.sqrt'))
    {
        cv <- get(paste0(what, '.cv'))
        
            for(var.name in names(cv)){
#                if(!grepl('fixed', var.name)){
#                    cv[[var.name]] %>% plot('image', main=var.name)
#                    cv[[var.name]] %>% plot('NN', main=var.name)
#                }
               if(what=="cov.sqrt"){ 
                cv[[var.name]] %>% plot('h', main=var.name,ylim=c(cov.min,cov.max))
               }
              if(what=="mean"){
                cv[[var.name]] %>% plot('h', main=var.name,ylim=c(mean.min,mean.max))
              }
            }
        }
}




### mean and covariance smoothing ###
if(compute.smooth.field){

    ### cross-validated parameters 1
    band.mode.mean <- 'effective.NN'  # type of adaptive bandwidth for the mean - alternative 'location.NN'
    band.mode.cov <- 'location.NN'    # type of adaptive bandwidth for the d-cov - alternative 'effective.NN'
    h1<-1.5 # bandwidth for the mean
    h2<-1  # bandwidth for the d-covariance
    mean.nn <- 300  # nearest neighbours for the mean
    cov.nn <- 32    # nearest neighbours for the covariance

    #prepared.data <- prepare_mfccs(mfcc.data, local.bnc, border.nodes)
    smooth.field.obs <- smooth_global(mfcc.data, local.bnc, border.nodes, mean.h=h1, mean.nn=mean.nn, cov.h=h2, cov.nn=cov.nn, mean.bandwidth.mode=band.mode.mean, cov.bandwidth.mode=band.mode.cov, fit.degree=fit.degree, min.angle=20, cov.smooth.type=cov.smooth.type)
    smooth.field.obs$pca$prcomps[1:2,,] <- -smooth.field.obs$pca$prcomps[1:2,,]


    ###### cross-validated parameters 2
    band.mode.mean <- 'location.NN'
    band.mode.cov <- 'location.NN'
    h1<-.5
    h2<-1
    mean.nn <- 14
    cov.nn <- 32 

    smooth.field.loc <- smooth_global(mfcc.data, local.bnc, border.nodes, mean.h=h1, mean.nn=mean.nn, cov.h=h2, cov.nn=cov.nn, mean.bandwidth.mode=band.mode.mean, cov.bandwidth.mode=band.mode.cov, fit.degree=fit.degree, min.angle=30, cov.smooth.type=cov.smooth.type)
    smooth.field.loc$pca$prcomps[1:2,,] <- -smooth.field.loc$pca$prcomps[1:2,,]

    save(smooth.field.obs, smooth.field.loc, file="aast_smooth.RData")
}


### pc plots ###    
if(show.pc.plots)
{ 
    load("aast_smooth.RData")

    ## nearest observations bandwidth
    op <- par(mfrow=c(3,2), oma=c(0,0,0,2))
    mai <- (par('mai'))
    mai[1] <- .3
    mai[2] <- .2
    for (pci in 1:3){ 
        par(mai=mai)
        ans <- smooth.field.obs %>% plot(what='mean.pc.smooth', pci=pci, res=res, asp=1, nlevel=nlevel, col=field.cols)
        lines(border.nodes,  col='black')
        contour(ans, add=TRUE, labcex=.3, lwd=.5)
        #  contour(ans, add=TRUE)
        mai.modif <- mai
        mai.modif[1] <- .51
        mai.modif[2] <- .7
        par(mai=mai.modif)
        smooth.field.obs$pca$prcomps[pci,,] %>% image.plot(y=1:10, main=paste0("PC ",pci, " loadings"), nlevel=nlevel, col=field.cols, xlab='time', ylab='MFCC')
    }
    par(op)

    ## nearest locations bandwidth
    op <- par(mfrow=c(3,2), oma=c(0,0,0,2))
    mai <- (par('mai'))
    mai[1] <- .3
    mai[2] <- .2
    for (pci in 1:3){ 
        par(mai=mai)
        ans <- smooth.field.loc %>% plot(what='mean.pc.smooth', pci=pci, res=res, asp=1, nlevel=nlevel, col=field.cols)
        lines(border.nodes,  col='black')
        contour(ans, add=TRUE, labcex=.3, lwd=.5)
        #  contour(ans, add=TRUE)
        mai.modif <- mai
        mai.modif[1] <- .51
        mai.modif[2] <- .7
        par(mai=mai.modif)
        smooth.field.loc$pca$prcomps[pci,,] %>% image.plot(y=1:10, main=paste0("PC ",pci, " loadings"), nlevel=nlevel, col=field.cols, xlab='time', ylab='MFCC')
    }
    par(op)

}
     

#### cov distances plots ###

if(sqrt.maps.compute){
    load("aast_smooth.RData")
    loc.coords <- prepared.data$loc.coords
    # generate and plot  the small images
    # WARNING: this is super slow
    image.list<-list()
    for (i in 1:length(location.list)){
        print(c(i, "/", length(location.list)) )
        k<-location.list[i]

        smooth.field2 <- smooth.field <- smooth.field.obs
        smooth.field2$loc.coords=loc.coords
        smooth.field2$border.nodes = border.nodes
        smooth.field2$mesh$nodes<-(smooth.field2$mesh$nodes-matrix(colMeans(border.nodes ),dim(smooth.field2$mesh$nodes)[1],2,byrow = TRUE))/10+matrix(loc.coords[k,],dim(smooth.field2$mesh$nodes)[1],2,byrow = TRUE)
        smooth.field2$loc.coords<-(smooth.field2$loc.coords-matrix(colMeans( border.nodes ),dim(smooth.field2$loc.coords)[1],2,byrow = TRUE))/10+matrix(loc.coords[k,],dim(smooth.field2$loc.coords)[1],2,byrow = TRUE)
        smooth.field2$border.nodes<-(smooth.field2$border.nodes-matrix(colMeans( border.nodes ),dim(smooth.field2$border.nodes)[1],2,byrow = TRUE))/10+matrix(loc.coords[k,],dim(smooth.field2$border.nodes)[1],2,byrow = TRUE)
        border.tmp<-( border.nodes - matrix(colMeans( border.nodes ), nrow(border.nodes),2,byrow = TRUE))/10+matrix(loc.coords[k,], nrow(border.nodes),2,byrow = TRUE)

        image.list[[i]]<-smooth.field2 %>% plot(what='dist.from.cov', from=k, pci=1, cov.std=FALSE, res=res, asp=1, cex.main=.75, nlevel=nlevel, col=field.cols,add=TRUE)
        lines(border.tmp,  col='black',lwd=1)
    }

    save(smooth.field, smooth.field2, location.list, image.list, file=paste0("cov_maps_sqrt.RData"))
}



if(l2.maps.compute){

    prepared.data <- prepare_mfccs(mfcc.data, local.bnc, border.nodes, min.angle=20, max.area=NA)


    ##### L2 smoothing of covariances ######
    smooth.field.l2 <- smooth_global_new(prepared.data, mean.h=1, mean.nn=300, mean.bandwidth.mode='effective.NN', fit.degree=0, cov.h=1, cov.nn=32, cov.bandwidth.mode='location.NN' , cov.smooth.type='l2')


    ## Computing pairwise distances; **slow**
    image.list <- list()
    for(i in 1:length(location.list)){
        print(c(i, "/", length(location.list)) )
        k <- location.list[i]
        smooth.field2<-smooth.field.l2
        smooth.field2$border.nodes = border.nodes
        smooth.field2$mesh$nodes<-(smooth.field2$mesh$nodes-matrix(colMeans(border.nodes ),dim(smooth.field2$mesh$nodes)[1],2,byrow = TRUE))/10+matrix(prepared.data$loc.coords[k,],dim(smooth.field2$mesh$nodes)[1],2,byrow = TRUE)
        smooth.field2$loc.coords<-(smooth.field2$loc.coords-matrix(colMeans( border.nodes ),dim(smooth.field2$loc.coords)[1],2,byrow = TRUE))/10+matrix(prepared.data$loc.coords[k,],dim(smooth.field2$loc.coords)[1],2,byrow = TRUE)
        smooth.field2$border.nodes<-(smooth.field2$border.nodes-matrix(colMeans( border.nodes ),dim(smooth.field2$border.nodes)[1],2,byrow = TRUE))/10+matrix(prepared.data$loc.coords[k,],dim(smooth.field2$border.nodes)[1],2,byrow = TRUE)
        border.tmp<-( border.nodes - matrix(colMeans( border.nodes ), nrow(border.nodes),2,byrow = TRUE))/10+matrix(prepared.data$loc.coords[k,], nrow(border.nodes),2,byrow = TRUE)
        image.list[[i]]<-smooth.field2 %>% plot(what='dist.from.cov', from=k, pci=1, cov.std=FALSE, res=200, asp=1, cex.main=.75, nlevel=nlevel.col, col=my.cols, add=FALSE)
    }

    save(smooth.field.l2, image.list, file=paste0('l2-maps-of-maps.RData'))
}



if(sqrt.vs.l2.maps){

    load(file=paste0('cov_maps_sqrt.RData')) %>% print
    image.list.sqrt <- image.list
    xrange <- smooth.field$mesh$nodes[,1] %>% range
    yrange <- smooth.field$mesh$nodes[,2] %>% range
    xseq0 <- seq(from=xrange[1], to=xrange[2], len=length(image.list.sqrt[[1]][['x']]))
    yseq0 <- seq(from=yrange[1], to=yrange[2], len=length(image.list.sqrt[[1]][['y']]))
    zlim.sqrt <- image.list.sqrt %>% lapply('[[', 'z2') %>% unlist %>% range(na.rm=TRUE)
    ## correction for the border
    xseq <- (xseq0 - mean(xseq0))*.99 + mean(xseq0)
    yseq <- (yseq0 - mean(yseq0))*.99 + mean(yseq0)
    for(i in 1:length(image.list.sqrt)){
        image.list.sqrt[[i]][['x']]  <- xseq
        image.list.sqrt[[i]][['y']]  <- yseq
        image.list.sqrt[[i]][['z2']]  <- image.list.sqrt[[i]][['z2']] / zlim.sqrt[2]
    }

    load( file=paste0('l2-maps-of-maps.RData')) %>% print
    image.list.l2 <- image.list; rm(image.list)
    zlim.l2 <- image.list.l2 %>% lapply('[[', 'z2') %>% unlist %>% range(na.rm=TRUE)
    for(i in 1:length(image.list.l2)){
        image.list.l2[[i]][['x']]  <- xseq
        image.list.l2[[i]][['y']]  <- yseq
        image.list.l2[[i]][['z2']]  <- image.list.l2[[i]][['z2']] / zlim.l2[2]
    }

    ## plot
    for(i in seq(along=location.list)){
        op <- par(mfrow=c(1,2), mar=c(0,2,4,5), oma=c(0,0,2,0))
        image.plot(x=xseq, y=yseq, image.list.sqrt[[i]]$z2,  nlevel=nlevel.col,  col=my.cols, bty='n', xaxt='n', yaxt='n', asp=1)
        title(main='Square-root Metric', line=0, cex.main=.8)
        lines(smooth.field$border.nodes, lwd=1)
        contour(image.list.sqrt[[i]], add=TRUE, zlim=zlim.sqrt/zlim.sqrt[2])
        image.plot(x=xseq, y=yseq, image.list.l2[[i]]$z2,  nlevel=nlevel.col,  col=my.cols, bty='n', xaxt='n', yaxt='n', asp=1)
        title(main='L2 Metric', line=0, cex.main=.8)
        lines(smooth.field$border.nodes, lwd=1)
        contour(image.list.l2[[i]], add=TRUE, zlim=zlim.l2/zlim.l2[2])
        title(main= row.names(smooth.field$loc.coords)[location.list[i]], outer=TRUE)
        par(op)
    }
}

if(geodesic.vs.euclidean.distance){
    smooth.field.geodesic <- smooth_global_new(prepared.data, mean.h=1, mean.nn=300, mean.bandwidth.mode='effective.NN', fit.degree=0, cov.h=1, cov.nn=32, cov.smooth.type='sqrt')

    prepared.data2 <- prepared.data
    prepared.data2$dist.mesh <- as.matrix(dist(prepared.data$mesh$nodes))
    smooth.field.euclidean <- smooth_global_new(prepared.data2, mean.h=1, mean.nn=300, mean.bandwidth.mode='effective.NN', fit.degree=0, cov.h=1, cov.nn=32, cov.smooth.type='sqrt')

    smooth.field.geodesic$pca$prcomps[1,,]  <- -smooth.field.geodesic$pca$prcomps[1,,]
    smooth.field.euclidean$pca$prcomps[1,,]  <- -smooth.field.euclidean$pca$prcomps[1,,]

    op <- par(mfrow=c(3,2), oma=c(0,0,0,2))
    mai <- (par('mai'))
    mai[1] <- .3
    mai[2] <- .2
    for (pci in 1:3){ 
        par(mai=mai)
        ans <- plot(smooth.field.geodesic, 'mean.pc.smooth', pci=pci, res=200, nlevel=nlevel.col, col=my.cols, asp=1)
        lines(border.nodes, col='black')
        contour(ans, add=TRUE)
        ans <- plot(smooth.field.euclidean, 'mean.pc.smooth', pci=pci, res=200, nlevel=nlevel.col, col=my.cols, asp=1)
        lines(border.nodes, col='black')
        contour(ans, add=TRUE)
    }
    par(op)
}

## output in 'simempirical-maps.tar'
if(constant.field.simulation){

    prepared.data <- prepare_mfccs(mfcc.data, local.bnc, border.nodes, min.angle=20, max.area=NA)
    set.seed(1)

    n.loc=nrow(prepared.data$loc.coords)

    nobs.per.loc.orig <- numeric(nrow(prepared.data$loc.coords))
    for(i in seq(nobs.per.loc.orig)){
        nobs.per.loc.orig[i] <- sum(prepared.data$local.bnc$placenamecleaned == row.names(prepared.data$loc.coords)[i])
    }
    loc.coords <- prepared.data$loc.coords
    nreps.per.loc <- nobs.per.loc.orig ## same number of observations per location as original data

    ## global mean and residuals
    mu <- prepared.data$mfcc.data %>% apply(2:3, mean)
    #mfcc.residuals <- sweep(prepared.data$mfcc.data, 2:3, mu)  # residuals from a global mean
    mfcc.residuals <- array(NA, dim(prepared.data$mfcc.data)) # residuals from a local mean
    {
        band.modes <- c('location.NN', 'effective.NN', 'fixed')
        mean.h <- 1
        mean.nn <- 300
        mean.bandwidth.mode <- band.modes[2]
        fit.degree <- 0

        ## compute mean field 
        mean.field <- mean_smooth(prepared.data$mfcc.data, cov.data=prepared.data$cov.data, prepared.data$obs.coords, loc.coords=loc.coords, output='estimation', h=mean.h, mesh=prepared.data$mesh, prepared.data$dist.mesh,  NN=mean.nn, bandwidth.mode=mean.bandwidth.mode, fit.degree=fit.degree) 
        for(i in 1:nrow(prepared.data$loc.coords)){
            nodei <- which(apply(mean.field$mesh$nodes, 1, identical, prepared.data$loc.coords[i,]))
            obsi <-  which(apply(prepared.data$obs.coords, 1, identical, prepared.data$loc.coords[i,]))
            mfcc.residuals[obsi,,] <- sweep(prepared.data$mfcc.data[obsi,,], 2:3, mean.field$mfcc[nodei,,])
        }
    }
    mfcc.residuals <- sweep(mfcc.residuals, 2:3, apply(mfcc.residuals, 2:3, mean)) # recenter the residuals

    n.residuals <- dim(mfcc.residuals)[1]

    ## generate obs.coords
    obs.coords <- array(NA, c(sum(nreps.per.loc), 2))
    begin.index <- 1
    for(i in seq(along=prepared.data$loc.coords[,1])){
        obs.coords[begin.index : sum(nreps.per.loc[1:i]),1] <-  prepared.data$loc.coords[i,1]
        obs.coords[begin.index : sum(nreps.per.loc[1:i]),2] <-  prepared.data$loc.coords[i,2]
        begin.index <- sum(nreps.per.loc[1:i])+1
    }

    for(sim.index in 1:100){

        print(c('Simulation', sim.index, '/', 100))

        ## generate mfcc data by resampling from the residuals
        mfcc.data <- array(NA, c(sum(nreps.per.loc), dim(mu)))
        for(i in 1:dim(mfcc.data)[1]){
            mfcc.data[i,,] <- mu + mfcc.residuals[sample(n.residuals, 1),,]
        }



        if(!file.exists(sprintf('simempirical-map-of-maps-%d.RData', sim.index))){ # only compute the cov field if it hasn't been done before
            file.create(sprintf('simempirical-map-of-maps-%d.RData', sim.index)) # create the .RData file (to allow parallel runs of this script)
            ## renormalize first mfcc
            mfcc.data[,,1] <- sweep(mfcc.data[,,1], 1, -mean(mfcc.data[,,1]) + rowMeans(mfcc.data[, , 1]))  # renormalizing cepstral coefficient 1 to have the same average everywhere

            ## compute cov.data
            n.sd <- rep(NA, dim(loc.coords)[1])
            cov.data <- array(NA, c(dim(loc.coords)[1], dim(mfcc.data)[2], dim(mfcc.data)[3], dim(mfcc.data)[3]))
            rm.index <- c()
            for (k in 1:dim(loc.coords)[1]){
                index <- which(apply(obs.coords, 1, identical, loc.coords[k,]))
                cov.data[k, ,,] <- cov_process(mfcc.data[index, ,])
                n.sd[k] <- length(index)
            }

            band.modes <- c('location.NN', 'effective.NN', 'fixed')
            mean.h <- 1
            mean.nn <- 300
            mean.bandwidth.mode <- band.modes[2]
            fit.degree <- 0

            ## compute mean field 
            mean.field.sim <- mean_smooth(mfcc.data, cov.data=cov.data, obs.coords, loc.coords=loc.coords, output='estimation', h=mean.h, mesh=prepared.data$mesh, prepared.data$dist.mesh,  NN=mean.nn, bandwidth.mode=mean.bandwidth.mode, fit.degree=fit.degree) 

            smooth.field.sim <- list(mesh=prepared.data$mesh, mfcc=mean.field.sim$mfcc, pca=mfccs_pc(prepared.data$mfcc.data, k=4)) ## projecting onto PCs of the original data
            #smooth.field.sim <- list(mesh=prepared.data$mesh, mfcc=mean.field.sim$mfcc, pca=mfccs_pc(mfcc.data, k=4))
            class(smooth.field.sim) <- 'smooth_field'
            #image.plot(smooth.field.sim$pca$prcomps[1,,])
            smooth.field.sim$pca$prcomps[1,,]  <- -smooth.field.sim$pca$prcomps[1,,]

            fit.degree <- 0
            cov.h <- 1
            cov.nn <- 32
            cov.smooth.type='sqrt'
            cov.bandwidth.mode <- band.modes[1]

            options(error=NULL)
            ## compute d-covariance field
            #cov.field <- list(cov=NA, taus=NA, effective.number.obs=NA)
            cov.field <- cov_smooth_sqrt(mfcc.data, obs.coords=obs.coords, loc.coords=loc.coords, output='estimation', mfcc.center=NULL, h=cov.h, mesh=prepared.data$mesh, n.sd=n.sd, prepared.data$dist.mesh, NN=cov.nn, fit.degree=fit.degree, bandwidth.mode=cov.bandwidth.mode, alpha=alpha)
            smooth.field.sim[['cov']]=cov.field$cov

            ## maps of maps
            location.names <- c("devon: newton abbot, uk" , "leicestershire: leicester, uk" , "west midlands: birmingham, uk" , "avon: bristol, uk" , "berkshire: sandhurst, uk" , "hampshire: andover, uk" , "staffordshire: newcastle under lyme, uk" , "essex: harlow, uk" , "kent: bickley, uk" , "wales, holywell, uk" , "hereford & worcester: redditch, uk" , "glasgow, uk" , "strathclyde: kilmarnock, uk" , "lancashire: morecambe, uk" , "norfolk: walpole st peter, uk" , "suffolk: ipswich, uk" , "greater london: westminster, uk" , "driffield, uk" , "dorset: poole, uk" , "lothian: edinburgh, uk" , "lothian: inverness, uk" , "south lanarkshire: lanark, uk" , "yorkshire, uk" , "shropshire, uk" , "northumberland: newcastle, uk" , "lincolnshire: lincoln, uk" , "northamptonshire: irthlingborough, uk" , "borders: galashiels, uk")
            location.list <- which(row.names(prepared.data$loc.coords) %in% location.names)

            loc.coords <- prepared.data$loc.coords
            ## Computing pairwise distances; **very slow**
            smooth.field2 <- smooth.field.sim
            image.list <- list()
            for(i in 1:length(location.list)){
                cat(c("---", i, "/", length(location.list), "---") )
                k <- location.list[i]
                smooth.field2<-smooth.field.sim
                smooth.field2$loc.coords=loc.coords
                smooth.field2$border.nodes = border.nodes
                smooth.field2$mesh$nodes<-(smooth.field2$mesh$nodes-matrix(colMeans(border.nodes ),dim(smooth.field2$mesh$nodes)[1],2,byrow = TRUE))/10+matrix(loc.coords[k,],dim(smooth.field2$mesh$nodes)[1],2,byrow = TRUE)
                smooth.field2$loc.coords<-(smooth.field2$loc.coords-matrix(colMeans( border.nodes ),dim(smooth.field2$loc.coords)[1],2,byrow = TRUE))/10+matrix(loc.coords[k,],dim(smooth.field2$loc.coords)[1],2,byrow = TRUE)
                smooth.field2$border.nodes<-(smooth.field2$border.nodes-matrix(colMeans( border.nodes ),dim(smooth.field2$border.nodes)[1],2,byrow = TRUE))/10++matrix(loc.coords[k,],dim(smooth.field2$border.nodes)[1],2,byrow = TRUE)
                border.tmp<-( border.nodes - matrix(colMeans( border.nodes ), nrow(border.nodes),2,byrow = TRUE))/10+matrix(loc.coords[k,], nrow(border.nodes),2,byrow = TRUE)
                image.list[[i]]<-smooth.field2 %>% plot(what='dist.from.cov', from=k, pci=1, cov.std=FALSE, res=200, asp=1, cex.main=.75, nlevel=nlevel.col, col=my.cols, add=FALSE)
            }
            save(smooth.field.sim, image.list, file=sprintf('simempirical-map-of-maps-%d.RData', sim.index))
        }
    }
}

if(constant.field.pc.plots){
    load(file='simempirical-map-of-maps-1.RData')
    ## plot mean field
    op <- par(mfrow=c(1,3), oma=c(0,0,0,2))
    mai <- (par('mai'))
    mai[1] <- .3
    mai[2] <- .2
    for (pci in 1:3){ 
        par(mai=mai)
        ans <- plot(smooth.field.sim, 'mean.pc.smooth', pci=pci, res=200, nlevel=nlevel.col, col=my.cols, asp=1)
        lines(border.nodes, col='black')
        contour(ans, add=TRUE, nlevels=7)
    }
    par(op)
}



if(plot.sim.contours){

    load(file=paste0('cov_maps_sqrt.RData')) %>% print
    image.list.bnc <- image.list
    xrange <- smooth.field$mesh$nodes[,1] %>% range
    yrange <- smooth.field$mesh$nodes[,2] %>% range
    xseq0 <- seq(from=xrange[1], to=xrange[2], len=length(image.list.bnc[[1]][['x']]))
    yseq0 <- seq(from=yrange[1], to=yrange[2], len=length(image.list.bnc[[1]][['y']]))
    zlim.bnc <- image.list.bnc %>% lapply('[[', 'z2') %>% unlist %>% range(na.rm=TRUE)
    ## correction for the border
    xseq <- (xseq0 - mean(xseq0))*.99 + mean(xseq0)
    yseq <- (yseq0 - mean(yseq0))*.99 + mean(yseq0)
    for(i in 1:length(image.list.bnc)){
        image.list.bnc[[i]][['x']]  <- xseq
        image.list.bnc[[i]][['y']]  <- yseq
        #    image.list.bnc[[i]][['z2']]  <- image.list.bnc[[i]][['z2']] / zlim.bnc[2]
    }

    image.list.sim <- list()
    ## load simulation images
    for(j in 1:100){
        print(j)
        load( file=sprintf('simempirical-map-of-maps-%d.RData',  j))
        image.list.sim[[j]] <- image.list; rm(image.list)
        ## zlim.sim <- image.list.sim %>% lapply('[[', 'z2') %>% unlist %>% range(na.rm=TRUE)
        for(i in 1:length(image.list.sim[[j]])){
            image.list.sim[[j]][[i]][['x']]  <- xseq
            image.list.sim[[j]][[i]][['y']]  <- yseq
            #    image.list.sim[[i]][['z2']]  <- image.list.sim[[i]][['z2']] / zlim.sim[2]
        }
    }

    levels=seq(.005, .045, by=.005)
    cols.transparent <- tim.colors(length(levels), alpha=100)
    cols <- tim.colors(length(levels))
    op <- par(mai=c(0,0,.8,0))
    for( ii in  1:length(image.list.bnc)){
        plot(smooth.field$border.nodes, lwd=1, type='l', asp=1, main=row.names(smooth.field$loc.coords)[location.list[ii]], xaxt="n",yaxt="n",xlab="",ylab="",bty="n")
        for(j in 1:100){
            contour(x=xseq, y=yseq, image.list.sim[[j]][[ii]]$z, add=T, drawlabels=F, levels=levels, col=cols.transparent, lwd=.3)
        }
        contour(x=xseq, y=yseq, image.list.bnc[[ii]]$z, add=TRUE, col=cols, lty=2, lwd=2, levels=levels)
        points(t(loc.coords[location.list[ii],]), pch="+", cex=2)
    }
    par(op)

}

if(compute.zscores){

    load(file='cov_maps_sqrt.RData') %>% print
    image.list.bnc <- image.list

    image.list.sim <- list()
    ## load simulation images
    for(j in 1:100){
        print(j)
        load( file=sprintf('simempirical-map-of-maps-%d.RData', j)) %>% print
        image.list.sim[[j]] <- image.list; rm(image.list)
        ## zlim.sim <- image.list.sim %>% lapply('[[', 'z2') %>% unlist %>% range(na.rm=TRUE)
    }

    zscores <- list()
    for( mapi in 1:length(image.list.bnc)){ # index of the map
        print(mapi)
        simi.array <- sapply(1:length(image.list.sim), function(x) image.list.sim[[x]][[mapi]]$z2, simplify='array') %>% aperm(c(3,1,2))
        simi.mean <- apply(simi.array, 2:3, mean)
        simi.sd <- apply(simi.array, 2:3, sd)
        zscores[[mapi]] <- list(z=(image.list.bnc[[mapi]]$z2 - simi.mean)/simi.sd, x=image.list.bnc[[mapi]]$x, y=image.list.bnc[[mapi]]$y)
    }

    save(zscores, file="zscores.empirical.RData")
}


if(plot.zscores){
    load(file="zscores.empirical.RData")

    xseq <- (xseq0 - mean(xseq0))*.99 + mean(xseq0)
    yseq <- (yseq0 - mean(yseq0))*.99 + mean(yseq0)

    ## plot maps of z-scores (separate) (not thresholded)
    my.cols2 <- c('#FFFFFFFF', my.cols[-(1)]) 
    zlim  <- zlim.bnc <- lapply(zscores, function(x) range(x$z, na.rm=TRUE) ) %>% unlist %>% range 
    breaks <- c(seq(from=zlim.bnc[1], 0, len=1), seq(from=0, to=zlim.bnc[2], len=nlevel.col+1)[-1])
    smooth.field2 <- smooth.field.sim
    border.nodes <- smooth.field$border.nodes
    loc.coords <- smooth.field$loc.coords
    op <- par(mai=c(0,0,.4,.8))
    for (mapi in 1:length(location.list)){
        k <- location.list[mapi]
        image.plot(x=xseq, y=yseq, z=zscores[[mapi]]$z, nlevel=nlevel,  col=my.cols2, zlim=zlim.bnc, useRaster = TRUE, breaks=breaks, xlim=range(smooth.field2$mesh$nodes[,1]), ylim=range(smooth.field2$mesh$nodes[,2]), xaxt="n",yaxt="n",xlab="",ylab="",bty="n",asp=1, main=row.names(loc.coords)[location.list[mapi]])
        lines(border.nodes,  col='black',lwd=2)
        points(t(loc.coords[k,]), pch='+', cex=2)
    }
    par(op)


    ## plot map of maps (not thresholded)
    my.cols2 <- c('#FFFFFFFF', my.cols[-(1)]) 
    zlim  <- zlim.bnc <- lapply(zscores, function(x) range(x$z, na.rm=TRUE) ) %>% unlist %>% range 
    breaks <- c(seq(from=zlim.bnc[1], 0, len=1), seq(from=0, to=zlim.bnc[2], len=nlevel.col+1)[-1])
    smooth.field2 <- smooth.field.sim
    border.nodes <- smooth.field$border.nodes
    loc.coords <- smooth.field$loc.coords
    op <- par(mai=c(0,0,0,.8))
    plot(range(smooth.field2$mesh$nodes[,1]), range(smooth.field2$mesh$nodes[,2]),col=0,xaxt="n",yaxt="n",xlab="",ylab="",bty="n",asp=1)
    lines(border.nodes,  col='black',lwd=2)
    for (mapi in 1:length(location.list)){
        k <- location.list[mapi]
        loc.coords2 <- (loc.coords - matrix(colMeans( border.nodes ), dim(loc.coords)[1], 2, byrow = TRUE))/10+matrix(loc.coords[k, ], dim(loc.coords)[1], 2, byrow = TRUE)
        border.tmp <- ( border.nodes  - matrix(colMeans( border.nodes ), nrow(border.nodes), 2, byrow = TRUE))/10+matrix(loc.coords[k, ], nrow(border.nodes), 2, byrow = TRUE)
        image.plot(zscores[[mapi]], nlevel=nlevel,  col=my.cols2, zlim=zlim, add=TRUE, useRaster = TRUE, breaks=breaks)
        lines(border.tmp,   col='black', lwd=1)
        points(t(loc.coords2[k,]), pch='+')
    }
    par(op)


    ## plot of two of the z-score maps
    zlim=lapply(zscores[c(18,19)], function(x) x$z) %>% unlist %>% range(na.rm=T)
    my.cols2 <- c('#FFFFFFFF', my.cols[-(1)]); nlevel.col2 <- length(my.cols2)
    breaks <- c(seq(from=zlim[1], 0, len=1), seq(from=0, to=zlim[2], len=nlevel.col2+1)[-1])
    op <- par(mfrow=c(1,2), oma=c(0,0,0,0), mai=c(0,.4, .4, .4))
    mapi=19 #9 or 19
    image.plot(x=xseq, y=yseq, z=zscores[[mapi]]$z, nlevel=nlevel,  col=my.cols2, zlim=zlim, useRaster = TRUE, breaks=breaks, bty='n', xaxt='n', yaxt='n', xlab='', ylab='', xlim=range(border.nodes[,1]), ylim=range(border.nodes[,2]), main=row.names(prepared.data$loc.coords)[location.list[mapi]], asp=1)
    contour(x=xseq, y=yseq, z=zscores[[mapi]]$z, add=TRUE)
    lines(border.nodes, lwd=1)
    points(t(prepared.data$loc.coords[location.list[mapi],]), pch=3, cex=5, lwd=1.3)
    print(row.names(prepared.data$loc.coords)[location.list[mapi]])
    mapi=18 #9 or 19
    image.plot(x=xseq, y=yseq, z=zscores[[mapi]]$z, nlevel=nlevel,  col=my.cols2, zlim=zlim, useRaster = TRUE, breaks=breaks, bty='n', xaxt='n', yaxt='n', xlab='', ylab='', xlim=range(border.nodes[,1]), ylim=range(border.nodes[,2]), main=row.names(prepared.data$loc.coords)[location.list[mapi]], asp=1)
    contour(x=xseq, y=yseq, z=zscores[[mapi]]$z, add=TRUE)
    lines(border.nodes, lwd=1)
    points(t(prepared.data$loc.coords[location.list[mapi],]), pch=3, cex=5, lwd=1.3)
    print(row.names(prepared.data$loc.coords)[location.list[mapi]])
    par(op)
    #print(range(zscores[[mapi]]$z, na.rm=T))
    #text((loc.coords[location.list,]), labels=as.character(1:28))

}


if(distance.vs.residual.compute){

    ##### Scatterplot of errors vs distance ###
    # get locations for each data point
    obs.index <- numeric(dim(prepared.data$mfcc.data)[1])
    for(i in seq(obs.index)){
        obs.index[i] = which(apply(prepared.data$loc.coords, 1, identical, prepared.data$obs.coords[i,]))
    }
    ## extract pairwise distances
    pairwise.distances <- prepared.data$dist.mesh[obs.index, obs.index] 

    ## extract residuals
    mfcc.data.local.means <- prepared.data %$% array(NA, c(nrow(loc.coords), dim(mfcc.data)[-1]))
    mfcc.residuals <- prepared.data %$% array(NA, dim(mfcc.data))
    attach(prepared.data)
    for(i in 1:nrow(loc.coords))
    {
        ## get which observations correspond to location i
        ind.tmp <- which(obs.coords[,1] == loc.coords[i,1] & obs.coords[,2] == loc.coords[i,2])
        apply( mfcc.data[ind.tmp,,], 2:3, mean ) -> mfcc.data.local.means[i,,]
        mfcc.residuals[ind.tmp,,] <-  sweep(mfcc.data[ind.tmp,,], 2:3,  mfcc.data.local.means[i,,]) 
    }
    anyNA(mfcc.residuals)
    sum((mfcc.residuals[1,,] - mfcc.residuals[2,,])^2)

    ## save pairwise residuals differences
    pairwise.residual.diff <- array(NA, dim(pairwise.distances))
    for(i in 1:(nrow(pairwise.distances)-1)){
        print(i)
        for(j in (i+1):nrow(pairwise.distances))
            pairwise.residual.diff[i,j] <- sum((mfcc.residuals[i,,]-mfcc.residuals[j,,])^2)
    }


    ## there are too many points to plot; subsample the points
    indx <- sample(nrow(pairwise.distances), 100)
    save(indx, pairwise.distances, pairwise.residual.diff, file='residuals.vs.distances.RData')
}

if(distance.vs.residual.plot){

    ## load data 
    load(file='residuals.vs.distance.RData') %>% print

    ## subsample points for fitting robust local linear regression
    subsamp <- sample(1:length(x), 1e5)
    lo.model <- lowess(x[subsamp], y[subsamp])
    

    ##  plot
    op <- par(mai=c(.4,.4, 0.1,0.1))
    plot(pairwise.distances[indx,], pairwise.residual.diff[indx,], type='p', pch='.', cex=.1, log='y', xlab='pairwise distance', ylab='distance between residuals')
    lines(lo.model, lwd=2)
    par(op)

}

if(covdist.vs.residual.compute){

    raw.dcovs <- array(NA, c(nrow(prepared.data$loc.coords), length(tsub), length(csub), length(csub)))
    for(i in 1:dim(raw.dcovs)[1]){
        print(i)
        raw.dcovs[i,,,] <- cov_process_sqrt(prepared.data$mfcc[which(apply(prepared.data$obs.coords, 1, identical,
                                                                           prepared.data$loc.coords[i,])),,] , centered=FALSE)
    }


    #index.dist.from <- 1 ## from which location should we take distances
    dist.from.location <- numeric(nrow(prepared.data$loc.coords))
    covdist.from.location <- numeric(nrow(prepared.data$loc.coords))
    covdists <- matrix(0, nrow(prepared.data$loc.coords),  nrow(prepared.data$loc.coords))
    dists <- matrix(0, nrow(prepared.data$loc.coords),  nrow(prepared.data$loc.coords))
    for(i in seq(along=prepared.data$loc.coords[,1])){
        #    dist.from.location[i] <- sqrt(sum((prepared.data$loc.coords[c(index.dist.from, i), ] %>% diff)^2))
        #    covdist.from.location[i] <- dist_cov_mfcc(raw.dcovs[index.dist.from,,,], raw.dcovs[i,,,], dist=sqrtdist,std=FALSE)
        print(i)
        for(j in 1:i){
            dists[i,j] <- sqrt(sum((prepared.data$loc.coords[c(i, j), ] %>% diff)^2))
            covdists[i,j] <- dist_cov_mfcc(raw.dcovs[i,,,], raw.dcovs[j,,,], dist=sqrtdist,std=FALSE)
        }
    }
    save(dists, covdists, file='covdist.vs.residual.RData')

}

if(covdist.vs.residual.plot){

    load(file='covdist.vs.residual.RData')

    lo.model <- lowess(x=dists[lower.tri(dists)], y=covdists[lower.tri(covdists)], iter=3)
    plot(dists[lower.tri(dists)], covdists[lower.tri(covdists)], cex=.6, pch='+', ylim=c(0, 1), xlab='geographical distance', ylab='pairwise d-cov distance')
    lines(lo.model, lwd=2)

}

if(duration.field.plot){

    durations.df <- data.frame(duration=aast.vowels$lengths/200/aast.vowels.bnc$duration, location=aast.vowels.bnc$placenamecleaned, sex=aast.vowels.bnc$sex, id=factor(aast.vowels.bnc$id), word=aast.vowels.bnc$word) ## duration is the relative duration of the vowel in the word


    durations.df %>% summary
    durations.df$sex[durations.df$sex == 'u'] <- NA
    durations.df <- mutate(durations.df, sex=factor(sex))
    durations.df <- filter(durations.df, !is.na(sex))
    #hist(durations.df$duration, 'fd')
    #hist(log(durations.df$duration), 'fd')
    durations.df <- mutate(durations.df, lduration=log(duration))

    library(lme4)

    ##
    model1 <- lmer(duration~location+sex+word+(1|id) , data=durations.df, REML=FALSE)
    #model1 <- lmer(duration~location+sex+(1|id) , data=durations.df, REML=FALSE)
    summary(model1)
    plot(model1) ## there is some heteroscedasticity

    ##
    model2 <- lmer(lduration~0+location+sex+word+(1|id) , data=durations.df, REML=FALSE)
    summary(model2)
    plot(model2) ## there is some heteroscedasticity
    model2@beta %>% str
    duration.per.location <- fixef(model2)
    duration.per.location  <- duration.per.location[1:128] # remove sex and word fixed effects
    library(stringr)
    ## cleaning names of locations
    names(duration.per.location) <- names(duration.per.location) %>% str_sub(9)
    duration.per.location.df <- data.frame(location.name=names(duration.per.location), lduration=duration.per.location, lon=NA, lat=NA)
    duration.per.location.df$duration  <- exp(duration.per.location.df$lduration) ## retransforming log durations to durations
    row.names(duration.per.location.df) <- 1:length(duration.per.location)
    ## retrieving lon-lat of locations
    load(sprintf("locations.RData"))
    for(i in 1:nrow(duration.per.location.df))
    {
        ind <- which(as.character(locations$location.names) == as.character(duration.per.location.df$location.name[i]))
        duration.per.location.df[i, c('lon', 'lat')] <- locations[ind, c('lon', 'lat')]
    }

    ## retrieve nodes that are inside Great-Britain
    which.inside <- which(1 == point.in.polygon(duration.per.location.df[, 'lon'], duration.per.location.df[, 'lat'] , border.nodes[, 1], border.nodes[, 2]))
    mesh <- create.MESH.2D(nodes=rbind(border.nodes, as.matrix(duration.per.location.df[which.inside, c('lon', 'lat')])), segments=border.segments)
    #plot(mesh, pch=20)
    mesh <- refine.MESH.2D(mesh, maximum_area = 1, minimum_angle=30)
    #plot(mesh, pch=20)
    ## Create the Finite Element basis 
    FEMbasis = create.FEM.basis(mesh)

    nlevel.col <- 64
    my.cols <- rev(heat.colors(nlevel.col))

    ## duration field smoothed with the graph distance
    prepared.data <- list(loc.coords=duration.per.location.df[which.inside, c('lon', 'lat')],
                          duration=duration.per.location.df[which.inside, 'duration'],
                          mesh=mesh, dist.mesh=NULL)
    prepared.data$obs.coords <- prepared.data$loc.coords
    rm(mesh)

    ## Smoothing with parameters from CV of smooth.field
    duration.smooth <- mean_smooth_duration(prepared.data, h=.5, NN=14, bandwidth.mode='location.NN', fit.degree=0)

    ## Plotting the results
    lambda=1e-10
    FEM = smooth.FEM.basis(locations=prepared.data$mesh$nodes, observations=duration.smooth$duration, FEMbasis=FEMbasis, lambda=lambda)
    op <- par(mar=c(0,0,4,2))
    ans <- FEM.image(FEM$fit.FEM, border.nodes=border.nodes, res=200, main=sprintf('Relative vowel duration'), data.location=NA, nlevels=nlevel.col, col=my.cols, xaxt='n', yaxt='n', bty='n', asp=1, xlab='', ylab='')
    lines(border.nodes)
    contour(ans, nlevels=8, add=TRUE)
    par(op)
}
